#!/usr/bin/perl
##############################################################################
#
#  ETS-RT -- Wolfenstein: Enemy Territory Real-Time Statistics
#  (c) Vladi Shabanski <cade@bis.bg> 2003
#  http://soul.datamax.bg/~cade/away/ets
#
#  Developed for play.evrocom.net server ET community
#  http://play.evrocom.net
#
#  This software is distributed under GPL license, see COPYING for details.
#
#  $Id: ets-rt.pl,v 1.8 2004/07/22 01:17:58 cade Exp $
#
##############################################################################
use Data::Dumper;
use IO::Socket::INET;
use IO::Select;
use IO::File;
use Storable;
use Fcntl;
use strict;

our $DEBUG = $ENV{'DEBUG'} || 0;

our $HERE = $ENV{'IMPOSTOR_HERE'} || $ENV{'ETS_HERE'} || 'play.evrocom.net';  # bind address of the host machine
our $LOG_CON_FILE = $ENV{ 'ETS_CON_FILE' };
our $WEB_PREFIX   = $ENV{ 'ETS_WEB_PREFIX' };
our $WEB_DIR      = $ENV{ 'ETS_WEB_DIR' };
our $OUTPUT_DIR   = $ENV{ 'ETS_OUTPUT_DIR' };
our $THEME        = $ENV{ 'ETS_THEME' } || "ets-default";
our $SERVER_IP    = $ENV{ 'ETS_SERVER_IP' };
our $SERVER_PWD   = $ENV{ 'ETS_SERVER_PWD' };
our $SEEN_FILE    = $ENV{ 'ETS_SEEN_FILE' };
our $QUOTES_FILE  = $ENV{ 'ETS_QUOTES_FILE' };
our $O_LOOP_TIME  = $ENV{ 'ETS_O_LOOP_TIME' } || 60; # 1 minute default
our $XPWINS_DIFF  = $ENV{ 'ETS_XPWINS_DIFF' } || 20; # minimum difference for xp wins theory calculation

$SERVER_IP .= ":27960" unless $SERVER_IP =~ /:\d+$/;

#my @data = split /\n+/, rcon('getstatus');
#my %data = split( /\\/, '.\.' . shift @data );
#print Dumper( \%data, \@data );
#exit;

our $revision = '$Revision: 1.8 $';
our $ets_version = $1 if $revision =~ /([\d\.]+)/;

### misc config ##############################################################

our $MOD = 0; # running mod 0 - etmain, 1 - etpro (autodetected)

### vars #####################################################################

our $SPREE_MIN  = 5;
our $SPREE_STEP = 5;
our %SPREE; # event/player -> count

our %BEST_SPREE;
our %BEST_SPREE_PLAYER;

our $SEEN = {};

##############################################################################

our %REFEREES = (
                '4345345345345345345340C13EDD6FDC' => 'cade',
                );

### messages #################################################################

our %MSG = (
  HEAD         => "^8***",
  MSG          => "^8[MSG]",
  SPREE        => "^7[PLAYER]^8 is in [SPREE] spree with ^7[COUNT]^8 (record: ^7[BEST_SP]^8)",
  SPREE_END    => "^7[PLAYER]'s^8 [SPREE] spree is ^7over^8 with ^7[COUNT]^8 in a row",
  SPREE_RECORD => "^7[PLAYER]^8 set [SPREE] spree ^7RECORD^8 with ^7[COUNT]^8 in a row",
  SPREE_BEST   => "[SPREE] spree ^7RECORD^8 is ^7[COUNT]^8 in a row scored by ^7[PLAYER]^8",
  SEEN         => "^7[PLAYER]^8 last seen ^7[TIME]^8 (^7[AGO]^8 ago)",
  SEEN_UNKNOWN => "^7[PLAYER]^8 is unknown",
  SEEN_MANY    => "^7[COUNT]^8 players found matching (^7[PLAYER]^8)",
           );


### ETS FIXME: THIS MUST BE MODULE ###########################################

my $LAST_DAYS    = $ENV{ 'ETS_LAST_DAYS' };
my $FLOOR        = $ENV{ 'ETS_FLOOR' } || 50;
my $SERVER_URL   = $ENV{ 'ETS_SERVER_URL' } || 'http://play.evrocom.net'; # this is credit line, please preserve it!

my $html_header = load_file ( $ENV{ 'ETS_HEADER_FILE_RT' } )
    || "<html>
        <head>
        <meta http-equiv='content-type' content='text/html; charset=[T:iso-8859-1]'>
        <meta http-equiv=refresh content=$O_LOOP_TIME>
        <title>Wolfenstein: Enemy Territory Statistics</title>
				<link href=#$THEME/styles/ets-default.css rel=stylesheet type=text/css>
				<style type=text/css></style>
        <head>
        <body>
        <center>
        <p>
        <h1 class='ets'>Wolfenstein: Enemy Territory Statistics</h1>
        <h2 class='ets'>[ENV:ETS_SERVER]</h2>
        <small>[T:This page has been generated on] [TIME]</small>
        <p>
        |
        <a href=!online.html>[T:online status]</a> |
        <a href=!tops.html>[T:top scores]</a> |
        <a href=!games.html>[T:last games]</a> |
        <a href=!players.html>[T:players (long!)]</a> |
        <a href=!clans.html>[T:clans]</a> |
        <a href=!kill.html>[T:kills/deaths/ratio]</a> |
        <a href=!classide.html>[T:sides/classes]</a> |
        <a href=!maps.html>[T:maps]</a> |
        <br>
        |
        <a href=!wep.html>[T:weapons]</a> |
        <a href=!covert.html>[T:covert operations]</a> |
        <a href=!kpm.html>[T:top kills per minute]</a> |
        <a href=!fsp.html>[T:frag spree]</a> |
        <a href=!noise.html>[T:noise]</a> |
        <a href=!jerks.html>[T:jerks]</a> |
        <a href=!waste.html>[T:time waste]</a> |
        <a href=!punks.html>[T:hall of shame]</a> |
        <br>
        |
        <a href=!p_sides.html>[T:sides played]</a> |
        <a href=!top_td.html>[T:'no friends' ladder]</a> |
        <a href=!harvest.html>[T:'harvest' ladder]</a> |
        <p>
        <table width=80%><td valign=top align=center>
        ";

my $html_footer = load_file ( $ENV{ 'ETS_FOOTER_FILE_RT' } );
if ( $html_footer ) {
    $html_footer =~ s/(\$\w+)/$1/gee; # variable expansion
} else {
    $html_footer = "</td></table><p><small>";
    $html_footer .= "Note: Statistics data for last $LAST_DAYS days only is used.<br>" if $LAST_DAYS > 0;
    $html_footer .= "Note: Limit is set to top $FLOOR entries (the rest are counted but not shown).<br>
       Thanks to all players at <a href=$SERVER_URL>$SERVER_URL</a>!<br>
       <a href=http://play.evrocom.net/cade/away/ets><b>ETS/RT $ets_version</b></a> by (c) <a href=mailto:cade\@bis.bg>&lt;cade\@bis.bg&gt;</a> 2003
       </small>
       </center>
       </body>
       </html>";
}

my $html_motd =               $ENV{ 'ETS_MOTD' }
                || load_file( $ENV{ 'ETS_MOTD_FILE' } )
                || "<h2 class='ets'>Welcome to Wolfenstein: Enemy Territory server statistics!</h2>";

### main() here ##############################################################

my %translation = split /(?<!\\)[=\n]+/, `cat $ENV{ETS_TRANSLATION}`
  if -e $ENV{ETS_TRANSLATION};

-e $LOG_CON_FILE or logger( "FATAL: cannot find $LOG_CON_FILE ($!)\n" );
my $FH = new IO::File( "/usr/bin/tail -f $LOG_CON_FILE |" ) or logger( "FATAL: cannot open $LOG_CON_FILE ($!)\n" );
my $flags = '';
fcntl( $FH, F_GETFL, $flags ) or logger( "ERROR: cannot get flags $LOG_CON_FILE ($!)" );
$flags |= O_NONBLOCK;
fcntl( $FH, F_SETFL, $flags ) or logger( "ERROR: cannot set flags $LOG_CON_FILE ($!)" );
#my $FH_S = IO::Select->new();
#$FH_S->add($FH);

my $break_main_loop = 0;

$SIG{ 'INT'  } = sub { $break_main_loop = 1; };
$SIG{ 'HUP'  } = sub { $break_main_loop = 1; };   
$SIG{ 'TERM' } = sub { $break_main_loop = 1; };   

$SEEN = retrieve( $SEEN_FILE ) if $SEEN_FILE and -e $SEEN_FILE;

# runtime vars
my %last_server;
my $last_game_time;
my %client;
my $time;
my $exit;

while(4)
  {
  $_ = <$FH>;
  last if $break_main_loop;
  online_status() if $OUTPUT_DIR and -d $OUTPUT_DIR;
  sleep(1) and next if $_ eq '';
  # logger( "DEBUG: CON: $_" );
  $time = localtime();

  if( /gamename: (\S+)/ ) { h_gamename( $1 ); next; }
  if( /^InitGame:\s*(.*)$/o )
    {
    $exit = 0;
    $last_game_time = $time;
    %last_server = split /\\/, "\\$1";
    # detect new etpro version game
    if ( $last_server{ 'gamename' } eq 'etpro' ) 
      {
      logger( "INFO: etpro mode detected" );
      $MOD = 1;
      }
    next;
    }
  if( /Exit:/o )
    {
    $last_server{ 'TIMELIMIT' } = /Timelimit/ ? 1 : 0;
    $exit = 1;
    next;
    }
  if( $exit and /^ShutdownGame/o )
    {
    %SPREE = ();
    %BEST_SPREE = ();
    %BEST_SPREE_PLAYER = ();
    next;
    }
  if( /^ClientUserinfoChanged: (\d+) (.+)/io )
    {
    # n\xwayBG\t\2\c\4\r\0\m\0000000\s\0000000\dn\chromosome\dr\0\w\10\lw\10\sw\41\mu\0\ref\0
    my $id = $1;
    my $data = $2;
    $client{ $id } = { split /\\/, $data };
    $SEEN->{ $client{ $id }{ 'n' } } = time();
    store( $SEEN, $SEEN_FILE ) if $SEEN_FILE;
    }
  if( /Kill: (\d+) (\d+) (\d+): (.+) killed (.+) by MOD_([A-Z_0-9]+)/io )
    {
    my $k = $4;
    my $d = $5;
    my $w = $6;
    
    h_spree_end( $d );
    h_spree( 'FRAG', $k );
    }
  if( /^([a-z]+_Pack): (\d+) (\d+)/io )
    {
    my $i = $1; # item
    my $f = $client{ $2 }{ 'n' }; # from 
    my $t = $client{ $3 }{ 'n' }; # to
    h_spree( "$i disp", $f );
    h_spree( "$i coll", $t );
    }
  if( /^(Medic_Revive): (\d+) (\d+)/io )
    {
    my $i = $1; # item
    my $f = $client{ $2 }{ 'n' }; # from 
    my $t = $client{ $3 }{ 'n' }; # to
    h_spree( "MEDIC REVIVE", $f );
    }
  if( /^say(buddy)?:\s+(.+?):\s+([\!\?].+)$/ )
    {
    my $buddy  = $1;
    my $player = $2;
    my $msg    = $3;
    
    # commands
    h_ref( $player ), next if $msg =~ /^\!REF/i;
    
    # queries
    h_best(), next if uc $msg eq '?BEST';
    h_seen( $1 ), next if uc $msg =~ /^\?SEEN\s+(\S+)/i;
    h_quote( $1 ), next if uc $msg =~ /^\?(\S+)/i;
    next;
    }
  }

### HANDLERS ################################################################
  
sub h_gamename
{
  my $game = shift;
}
  
sub h_spree
{
  my $event  = uc shift;
  my $player =    shift;
  return unless $player;
  
  logger( "DEBUG: spree $event, $player, " . $SPREE{ $player }{ $event } ) if $SPREE{ $player }{ $event } > 0;
  
  $SPREE{ $player }{ $event }++;
  my $sp = $SPREE{ $player }{ $event };
  return if $sp < $SPREE_MIN;
  return if $sp % $SPREE_STEP;
  if( $sp > $BEST_SPREE{ $event } )
    {
    $BEST_SPREE{ $event } = $sp;
    $BEST_SPREE_PLAYER{ $event } = $player;
    et_print( 'SPREE_RECORD', SPREE => $event, PLAYER => $player, COUNT => $sp )
    }
  else
    {
    et_print( 'SPREE', SPREE => $event, PLAYER => $player, COUNT => $sp, BEST_SP => $BEST_SPREE{ $event } );
    }
}

sub h_spree_end
{
  my $player = shift;
  return unless $player;

  for my $event ( keys %{ $SPREE{ $player } } )
    {
    my $sp = $SPREE{ $player }{ $event };
    next if $sp < $SPREE_MIN;
    if( $sp > $BEST_SPREE{ $event } )
      {
      $BEST_SPREE{ $event } = $sp;
      $BEST_SPREE_PLAYER{ $event } = $player;
      et_print( 'SPREE_RECORD', SPREE => $event, PLAYER => $player, COUNT => $sp )
      }
    else
      {
      et_print( 'SPREE_END', SPREE => $event, PLAYER => $player, COUNT => $sp );
      }  
    }
  delete $SPREE{ $player };
}

sub h_best
{
  for my $event ( keys %BEST_SPREE )
    {
    my $player = $BEST_SPREE_PLAYER{ $event };
    my $sp     = $BEST_SPREE{ $event };
    et_print( 'SPREE_BEST', SPREE => $event, PLAYER => $player, COUNT => $sp )
    }
}

sub h_seen
{
  my $player = lc shift;
  $player =~ s/\^.//g;
  
  my $pl;
  my $tm;
  my $c;
  while( my ( $k, $v ) = each %$SEEN )
    {
    my $kk = $k;
    $kk =~ s/\^.//g;
    $kk = lc $kk;
    next if index( $kk, $player ) < 0;
    $c++;
    $pl = $k;
    $tm = $v;
    }
  et_print( 'SEEN_MANY', PLAYER => $player, COUNT => $c ) if $c > 1;
  et_print( 'SEEN_UNKNOWN', PLAYER => $player ) if $c == 0;
  if( $c == 1 )
    {
    my $ago = time() - $tm;
    $ago = int($ago/60); # now in minutes
    $ago = sprintf "%dd %dh %dm", int(   $ago / ( 24 * 60 ) ), 
                                  int( ( $ago % ( 24 * 60 ) ) / 60 ),
                                  int( ( $ago % ( 24 * 60 ) ) % 60 );
    et_print( 'SEEN', PLAYER => $pl, TIME => scalar(localtime($tm)), AGO => $ago );
    }
}

our @QUOTES;
sub h_quote
{
  my $word = shift;
  @QUOTES = `cat $QUOTES_FILE` unless @QUOTES;
  
  my $border = $QUOTES[-1];
  while(4)
    {
    my $q = shift @QUOTES;
    push @QUOTES, $q;
    last if $q eq $border;
    if( $q =~ s/(\Q$word\E)/^7$1^8/gi )
      {
      et_print( 'MSG', MSG => $q );
      last;
      }
    }
}

sub h_ref
{
  my $player  = shift;
  $player =~ s/\^.//g;

  # return unless $MOD == 1;
  
  my @rcon = split /\n/, rcon( "pb_sv_plist" );
  my $found = 0;
  for( @rcon )
    {
    s/^PunkBuster Server:\s*//;
    next if /^(End of )?Player List/;
    my @pb = split /\s+/, $_, 9;
    # 0 [1]
    # 1 [f112ad6c41a48ada6b4cfc2e76fac9c1(?)]
    # 2 [193.200.15.141:29111]
    # 3 [OK]
    # 4 [1]
    # 5 [0.0]
    # 6 [0]
    # 7 [(W)]
    # 8 [[ZaZ]KserkS]
    $pb[1] =~ s/\(.*?\)$//;
    $pb[2] =~ s/:\d*$//;
    $pb[8] =~ s/\^.//g;
    if( $player eq $pb[8] and $REFEREES{ uc $pb[1] }  )
      {
      rcon( "ref referee $pb[8]" );
      return;
      }
    }
}

### ONLINE STATUS ###########################################################

our $O_LAST_MAP;
our $O_LAST_TIME;
our $O_LAST_OK;
sub online_status
{
  return if $O_LAST_OK and time() % $O_LOOP_TIME;
  
  my $rcon = rcon('getstatus');
  $O_LAST_OK = $rcon ? 1 : 0;
  return unless $O_LAST_OK;
  my @srv = split /\n+/, $rcon;
  my %srv = split( /\\/, '.\.' . shift @srv );

  # players -----------------------------------------------------
  
  my @pdata;

  my %side;
  
  if( $srv{ 'Players_Allies' } and  $srv{ 'Players_Axis'   } )
    {
    $side{ $_ } = 'Ally' for split /\s+/, $srv{ 'Players_Allies' };
    $side{ $_ } = 'Axis' for split /\s+/, $srv{ 'Players_Axis'   };
    }
  elsif( $srv{ 'P' } )
    {
    my $c = 1;
    $side{ $c++ } = ( '????','Axis','Ally','Spct' )[$_] for split //, $srv{ 'P' };
    }
  
  my %pl = ( 'Ally' => [], 'Axis' => [], 'Spct' => [] );

  my %xp;
  my $c = 1;
  for( @srv )
    {
    chomp();
    next unless /\S/;
    my ( $xp, $ping, $name ) = split /\s+/, $_, 3;
    $name =~ s/"//g; # "
    my $s = $side{ $c } || 'Spct';
    $xp{ $s } += $xp;
    push @{ $pl{ $s } }, [ $ping, $xp, substr( $name, 0, 25 ) ];
    $c++;
    }

  #for my $s ( ( 'Axis', 'Ally', 'Spct' ) )
  #  {
  #  @{ $pl{ $s } } = reverse sort @{ $pl{ $s } };
  #  }

  push @pdata, [ 'HEAD', '[>5%]Ping', '[>5%]XP', '[<23%]Axis', 
                         '[>5%]Ping', '[>5%]XP', '[<23%]Ally',
                         '[>5%]Ping', '[>5%]XP', '[<24%]Spct', ];
  #print Dumper( \%pl );
  while(4)
    {
    my @line;
    my $c;
    for my $s ( ( 'Axis', 'Ally', 'Spct' ) )
      {
      my $ar = shift( @{ $pl{ $s } } );
      #print Dumper( $pl{ $s });
      push @line, $ar ? @$ar : ( '', '', '' );
      $c++ if $ar;
      }
    last unless $c;
    push @pdata, [ '', @line ];
    }
  push @pdata, [ 'FOOT', '', '(SUM)', '', 
                         '', '(SUM)', '',
                         '', '(SUM)', '', ];

  # server data -------------------------------------------------

  my $map = $srv{ 'mapname' };
  if( $O_LAST_MAP ne $map )
    {
    $O_LAST_MAP = $map;
    $O_LAST_TIME = time();
    }

  my $eta = 'unknown';
  if( $O_LAST_TIME )
    {
    $eta = int( $srv{'timelimit'} - ( time() - $O_LAST_TIME )/60 );
    $eta = "overtime/unknown" if $eta < 0;
    }

  my $gt = $srv{ 'g_gametype' };
  my $gts = 'unknown';
  $gts = 'Campaign' if $gt == 4;
  $gts = 'Stopwatch' if $gt == 3;
  $gts = 'Last Man Standing' if $gt == 5;
  $gts = 'Single Map Objective' if $gt == 2;

  my $winner = 'unknown';
  my $axp = $xp{ 'Axis' };
  my $ixp = $xp{ 'Ally' };
  my $max = $axp > $ixp ? $axp : $ixp;
  my $dif = $max > 0 ? ( 100 * abs( $axp - $ixp ) ) / $max : 0;
  if( $dif > $XPWINS_DIFF )
    {
    $winner = 'Axis'   if $axp > $ixp;
    $winner = 'Allies' if $axp < $ixp;
    };
  $dif = sprintf "%.2f", $dif;  


  my @data;
  push @data, [ 'HEAD', '[>50%]Name', '[<50%]Value' ];
  push @data, [ '', 'Host name',       $srv{'sv_hostname'} ];
  push @data, [ '', 'Version',         $srv{'version'} ];
  push @data, [ '', 'Game name (mod)', $srv{'gamename'} ];
  push @data, [ '', 'URL',             $srv{'URL'} ] if $srv{'URL'};
  push @data, [ '', 'Map',             $srv{'mapname'} ];
  push @data, [ '', 'Game type',       $gts ];
  push @data, [ '', 'Timelimit',       $srv{'timelimit'} . " min" ];
  push @data, [ '', 'Map started at',  scalar(localtime($O_LAST_TIME)) ];
  push @data, [ '', 'Finish ETA in',   "$eta min" ];
  push @data, [ '', 'XP Difference',   "$dif%" ];
  push @data, [ '', 'Probable winner', $winner ];
  
  my $text;

  my $scr = next_screen( $map );
  $text .= html_table( \@data );

  html_save( 'online', make_columns( "<img src=#screenshots/$scr width width=400 height=300>", $text ),
                       "<br>",
                       html_table( \@pdata )  );
}

our @SCREENS;

sub next_screen
{
  my $map = shift;
  
  if( ! @SCREENS )
    {
    @SCREENS = glob "$OUTPUT_DIR/screenshots/*.jpg";
    s/^(.+?)([^\/]+)$/$2/ for @SCREENS;
    }
  
  my $border = $SCREENS[-1];
  while(4)
    {
    my $q = shift @SCREENS;
    push @SCREENS, $q;
    last if $q eq $border;
    return $q if $q =~ /\Q$map\E/i;
    }
  return 'no-screenshot.jpg';  
}

### UTILITIES ###############################################################

sub et_print
{
  my $msg_id = uc shift;
  my %opt    = @_;
  
  my $say  = $MOD ? 'qsay' : 'say';
  my $head = $MSG{ 'HEAD' };
  my $msg = $MSG{ $msg_id };
  
  $msg =~ s/\[([^\]]+)\]/$opt{$1}/g;
  
  rcon( "$say $head $msg" );
}

sub rcon
{
  my $command = shift;
  
  return logger( "RCON DUMMY: $command" ) if $DEBUG;
  
  my $sock = IO::Socket::INET->new( PeerAddr  => $SERVER_IP,
                                    Proto     => 'udp',
                                    LocalAddr => $HERE,
                                    Broadcast => '1' ) or 
  logger( "ERROR: cannot connect $SERVER_IP" );
      
  my $sel = IO::Select->new();
  $sel->add($sock);
  
  logger( "status: rcon <<< $command" );
  if( $command eq 'getstatus' )
    {
    $sock->send( "\xff\xff\xff\xffgetstatus\n" );
    }
  elsif( $command eq 'getinfo' )
    {
    $sock->send( "\xff\xff\xff\xffgetinfo\n" );
    }
  else
    {  
    $sock->send( "\xff\xff\xff\xffrcon $SERVER_PWD $command" );
    }
  my $data;
  while( my @ready = $sel->can_read( 1 ) )
    {
    $sock->recv( $_, 4096 );
    s/\xff+(.*?)\n//;
    $data .= $_;
    };
  $data =~ s/\^.//g;
  
  my @data = split /\n+/, $data;
  my $data1 = $data[0];
  $data1 .= "..." if @data > 1;
  logger( "status: rcon >>> $data1" );
  return $data;
}

sub logger
{
  my $msg = shift;
  my $tm = scalar localtime();
  print STDERR "[$tm] $msg\n";
  exit(1) if $msg =~ /^FATAL:/;
}

### ETS FIXME: THIS MUST BE MODULE ###########################################

sub make_columns
{
  my $text;
  
  my $c = @_;
  my $p = $c > 0 ? int( 100 / @_ ) : 100;
  
  $text .= "<table cellspacing=5 width=100% class=etscolumns><tr><td width=p% valign=top>";
  $text .= join "</td><td width=$p% valign=top>", @_;
  $text .= "</td></tr></table>";
  
  return $text;
}

sub html_table_cmp
{
  my $aa = shift;
  my $bb = shift;
  my $ta = shift;
  my $tb = shift;
  my $sa = shift;
  
  return -1 if $ta eq 'HEAD';
  return  1 if $tb eq 'HEAD';
  return  1 if $ta eq 'FOOT';
  return -1 if $tb eq 'FOOT';
  
  $aa = $3 if $aa =~ /^(\[([^\]]+)\])?(.+)(\/([ip]))?$/o;
  $bb = $3 if $bb =~ /^(\[([^\]]+)\])?(.+)(\/([ip]))?$/o;
  
  return $sa ? $aa cmp $bb : $bb <=> $aa;
}

sub html_table
{
  my $ar = shift;
  my %opt = @_;
  
  my $text;
  $text .= "<table class=ets width=100% cellspacing=1 cellpadding=5>\n";
  my @or; # options row
  my @ft; # format type
  my $r = -1; # row
  
  my @ar = @$ar;
  
  my $floor = $opt{ 'FLOOR' } || $FLOOR;
  
  my $oc = $opt{ 'SORT' } || $opt{ 'SORTA' }; # sort/order column
  @ar = sort { html_table_cmp( $a->[$oc], $b->[$oc], $a->[0], $b->[0], $opt{ 'SORTA' } ) } @ar if $oc > 0;
  
  for my $rr ( @ar )
    {
    $r++ if $rr->[0] eq '';
    if( $floor > 0 )
      {
      next if $rr->[0] eq 'COMT' and $r >= $floor;
      next if $rr->[0] eq ''     and $r >= $floor;
      }
    my $c = -1; # column
    my $class = tr_class( $r );
    @or = @$rr and next if $rr->[0] eq 'OPT';
    $class = 'hh' if $rr->[0] eq 'HEAD'; # header
    $class = 'ff' if $rr->[0] eq 'FOOT'; # footer
    $class = 'cc' if $rr->[0] eq 'COMT'; # comments
    $text .= "<tr class=$class>\n";
    for( @$rr )
      {
      my $d = $_;
      $c++;
      next if $c == 0;
      
      if( $d =~ s/^\[([^\]]+)\]//o )
        {
        my $p = $1; # params
        my $o; # opts
        $o .= " align=right" if $p =~ />/o;
        $o .= " width=$1"    if $p =~ /(\d+%)/o;
        $or[$c] = $o;
        $ft[$c] = lc $1 if $p =~ /([ip])$/io;
        #print STDERR "TABLE OPT: $p --> $o\n";
        #$text .= "TABLE OPT: $p --> $o<br>";
        }
      my $f;
      $f = $ft[$c] if $rr->[0] ne 'HEAD' and $rr->[0] ne 'FOOT';
      $f = lc $1 if $d =~ s/\/([ip])$//io; # [i]nt, [p]ercent
      $d = table_sum( $ar, $c ) if $d eq '(SUM)';
      $d = table_ave( $ar, $c ) if $d eq '(AVE)';
      $d = table_med( $ar, $c ) if $d eq '(MED)';
      
      $d = num_fmt( int $d ) if $f eq 'i';
      $d = $d > 0 ? sprintf "%.2f%%", $d : '' if $f eq 'p';
      
      $d = $r + 1 if $d eq '(ROW)';
      $d = "<span class=hitext>$d</div>" if $oc > 1 and $c == $oc and ! $$rr[0];
      $text .= "<td class=etstd $or[$c]>$d</td>\n";
      }
    $text .= "</tr>\n";
    }
  $text .= "</table>\n";
  return $text;
}

sub table_sum
{
  my $ar = shift;
  my $col = shift;
  my $sum = 0;
  for my $rr ( @$ar )
    {
    next if $rr->[0] ne ''; # skip HEAD,FOOT
    $sum += $rr->[$col];
    }
  return $sum;
}

sub table_ave
{
  my $ar = shift;
  my $col = shift;
  my $sum = 0;
  my $cnt = 0;
  for my $rr ( @$ar )
    {
    next if $rr->[0] ne ''; # skip HEAD,FOOT
    $sum += $rr->[$col];
    $cnt++;
    }
  return $cnt > 0 ? $sum / $cnt : 0;
}

sub table_med
{
  my $ar = shift;
  my $col = shift;
  my @sum;
  for my $rr ( @$ar )
    {
    next if $rr->[0] ne ''; # skip HEAD,FOOT
    push @sum, $rr->[$col];
    }
  @sum = sort { $b <=> $a } @sum;  
  return $sum[ @sum / 2 ];
}

sub time_fmt
{
  my $data = shift;
  my $d = int($data / (24*60*60));
  my $h = int(($data % (24*60*60)) / (60*60));
  my $m = int(($data % (60*60)) / 60);        
  my $s = $data % 60;                 
  return sprintf "%6dd %02dh %02dm %02ds", $d, $h, $m, $s;
}

sub num_fmt
{
  my $s = shift;
  $s = reverse $s;
  1 while $s =~ s/(\d\d\d)(\d)([^']*)$/$1'$2$3/;
  $s = reverse $s;
}

sub html_save
{
  my $file = shift;
  $file = "$OUTPUT_DIR/$file.html";
  my $text = join '', $html_header, @_, $html_footer;
  my $time = scalar localtime();
  $text =~ s/\[TIME\]/$time/g;
  $text =~ s/\[ENV:([^\]]+)\]/$ENV{$1}/g;
  $text =~ s/href=!/href=$WEB_PREFIX/gio;
  $text =~ s/href=#/href=$WEB_DIR/gio;
  $text =~ s/src=#/src=$WEB_DIR/gio;
  $text =~ s/\[T\:([^\]]*)\]/$translation{$1}||$1/gie; # \: is to avoid get_trans.pl
  save_file( $file, $text ) or die "$0: cannot save html file: $file\n";
  #print STDERR "file $file written ok\n";
}

sub tr_class
{
  $_[0] % 2 ? 'r1' : 'r2';
}

sub hash_max
{
  my $hr = shift;
  my $max_k = undef;
  my $max_v = undef;
  while( my ( $k, $v ) = each %$hr )
    {
    if( $max_v eq '' or $v > $max_v )
      {
      $max_k = $k;
      $max_v = $v;
      next;
      }
    }
  #print STDERR "$max_k, $max_v\n";
  return ( $max_k, $max_v );
}

sub hash_sum
{
  my $hr = shift;
  my $sum = 0;
  while( my ( $k, $v ) = each %$hr )
    {
    next if $k =~ /^!/;
    $sum += $v;
    }
  return $sum;
}

sub load_file
{
  my $file = shift;
  my $i;
  open( $i, $file ) or return undef;
  local $/ = undef;
  my $s = <$i>;
  close $i;
  return $s;
}

sub save_file
{
  my $file = shift;
  my $o;
  open( $o, ">$file" ) or return 0;
  print $o @_;
  close $o;
  return 1;
}

#############################################################################

=pod

$VAR1 = {
          'protocol' => '83',
          'sv_maxRate' => '25000',
          'g_minGameClients' => '8',
          'sv_allowAnonymous' => '0',
          'mapname' => 'raiders_b2',
          'g_axismaxlives' => '0',
          'g_needpass' => '0',
          'sv_hostname' => 'play.evrocom.net',
          'URL' => 'http://play.evrocom.net/',
          'sv_punkbuster' => '1',
          'g_alliedmaxlives' => '0',
          'mod_version' => '3.0.13',
          'g_balancedteams' => '1',
          'g_antilag' => '1',
          'g_maxlivesRespawnPenalty' => '0',
          'g_medicChargeTime' => '45000',
          'sv_floodProtect' => '0',
          'g_engineerChargeTime' => '30000',
          'g_friendlyFire' => '1',
          'g_bluelimbotime' => '15000',
          'g_redlimbotime' => '25000',
          'g_soldierChargeTime' => '20000',
          'g_maxGameClients' => '0',
          'sv_privateClients' => '0',
          'timelimit' => '30',
          'g_covertopsChargeTime' => '30000',
          'g_voteFlags' => '0',
          'sv_maxPing' => '200',
          'sv_minguidage' => '0',
          'g_maxlives' => '0',
          'sv_maxclients' => '16',
          '.' => '.',
          'voteFlags' => '65535',
          'g_heavyWeaponRestriction' => '20',
          'version' => 'ET 2.56 linux-i386 Sep 10 2003',
          'P' => '-21---11222-12-1',
          'gamename' => 'etpro',
          'g_gametype' => '4',
          'sv_minPing' => '0',
          'mod_url' => 'http://etpro.anime.net/',
          'g_LTChargeTime' => '40000'
        };

=cut
